-- Audio Effects Unit: Hierarchical

-- 1 to 4 DEMUX for beginning of audio effects unit
library ieee;
use ieee.std_logic_1164.all;
entity demux_1_to_4 is
  port (MUX_CONTROL: in std_logic_vector(1 downto 0); -- Control logic for MUX/DEMUX
        PIN: in std_logic; --16 bit one-channel stereo audio sample
        S: out std_logic_vector(0 to 3));
end demux_1_to_4;

architecture function_table_ws of demux_1_to_4 is
begin
  with MUX_CONTROL select  
    PIN  => S(0) when "00",         -- used to be I
         S(1) when "01",
         S(2) when "10",
         S(3) when "11",
         'X' when others;
   S0 <= S(0); -- Normal Operation
   S1 <= S(1); -- Flanging
   S2 <= S(2); -- Echo
   S3 <= S(3); -- Reverb/Vibrato
end function_table_ws;
-- COMPLETE 1 to 4 DEMUX


-- FLANGING UNIT ATTACHED to O(1) and E(1)

  -- 0 to 47 counter (~1ms) to act as a clock with period of ~ 1ms
  library ieee;
  use ieee.std_logic_1164.all;
  entity 1_ms_clock is
     port (AU_LRCK : in std_logic;      -- takes in fast clock
               MSC : out std_logic);    -- outputs clock with T = 1ms
  end 1_ms_clock;

  architecture behavioral of 1_ms_clock is
    begin
    count = 0;
    up <= 0;
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      if (count + 1)%98 = '0' then
        count = 0;
        up <= not up;
          else
            up <= 1;
          endif;
      endif;        
    endif;
  end process;
    MSC <= up;
  end behavioral;

  -- 1 to 10 to 1 counter controlling select on 1_to_10 DEMUX, on 1 ms clock
  library ieee;
  use ieee.std_logic_1164.all;
  use IEEE.STD_LOGIC_ARITH.ALL;  
  entity bouncer_1_to_10 is
    port (MSC : in std_logic; -- takes MSC (1 ms clock), not AU_BCLK
          BCR : out std_logic); -- bounce signal
  end bouncer_1_to_10;
  
  architecture behaviorial of bouncer_1_to_10 is
    up <= 0;
    count <= 1;
  begin
  process (MSC)
  begin
    if (MSC'event and (MSC = '1')) then
      if up = '0' then
        count = count + 1;
      else
        count = count - 1;
    endif;
    if count = '1' or count = '10' then
      up <= not up;
    endif;
  end process;
    BCR <= count;
  end behavioral;

  -- ======================================================
  --    FLANGER DELAY ELEMENTS
  -- All of these use the 98 KHz clock
  
  -- FOR ELEMENT N (N!=1):
  --    input is TN
  --    output is TNO
  -- FOR ELEMENT 1
  --    input is S1 from effect select DEMUX
  -- ======================================================
    -- ~1 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_1 is 
      port(AU_LRCK, S1 : in std_logic;
      QT1  : out std_logic_vector(97 downto 0);
      T1O : out std_logic);
    end srg_1;
  
    architecture behavioral of srg_1 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & S1;
      endif;
    end process;
      QT1 <= shift;
      T1O <= shift(97);
    end behavioral;
    
    -- 2 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_2 is 
      port(AU_LRCK, T1O : in std_logic;
      QT2  : out std_logic_vector(97 downto 0);
      T2O : out std_logic);
    end srg_2;
  
    architecture behavioral of srg_2 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T2;
      endif;
    end process;
      QT2 <= shift;
      T2O <= shift(97);
    end behavioral;
    
    -- 3 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_3 is
      port(CLK, T2O : in std_logic;
      QT3  : out std_logic_vector(97 downto 0);
      T3O : out std_logic);
    end srg_3;
  
    architecture behavioral of srg_3 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T3;
      endif;
    end process;
      QT3 <= shift;
      T3O <= shift(97);
    end behavioral;
    
    -- 4 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_4 is
      port(AU_LRCK, T3O : in std_logic;
      QT4  : out std_logic_vector(97 downto 0);
      T4O : out std_logic);
    end srg_4;
  
    architecture behavioral of srg_4 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T4;
      endif;
    end process;
      QT4 <= shift;
      T4O <= shift(97);
    end behavioral;
    
    -- 5 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_5 is
      port(AU_LRCK, T4O : in std_logic;
      QT5  : out std_logic_vector(97 downto 0);
      T5O : out std_logic);
    end srg_5;
  
    architecture behavioral of srg_5 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T5;
      endif;
    end process;
      QT5 <= shift;
      T5O <= shift(97);
    end behavioral;
    
    -- 6 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_6 is
      port(AU_LRCK, T5O : in std_logic;
      QT6  : out std_logic_vector(97 downto 0);
      T6O : out std_logic);
    end srg_6;
  
    architecture behavioral of srg_6 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T6;
      endif;
    end process;
      QT6 <= shift;
      T6O <= shift(97);
    end behavioral;
    
    -- 7 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_7 is
      port(AU_LRCK, T6O : in std_logic;
      QT7  : out std_logic_vector(97 downto 0);
      T7O : out std_logic);
    end srg_7;
  
    architecture behavioral of srg_7 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T7;
      endif;
    end process;
      QT7 <= shift;
      T7O <= shift(97);
    end behavioral;
    
    -- 8 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_8 is
      port(AU_LRCK, T7O : in std_logic;
      QT8  : out std_logic_vector(97 downto 0);
      T8O : out std_logic);
    end srg_8;
  
    architecture behavioral of srg_8 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T8;
      endif;
    end process;
      QT8 <= shift;
      T8O <= shift(97);
    end behavioral;
    
    -- 9 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_9 is
      port(AU_LRCK, T8O : in std_logic;
      QT9  : out std_logic_vector(97 downto 0);
      T9O : out std_logic);
    end srg_9;
  
    architecture behavioral of srg_9 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T9;
      endif;
    end process;
      QT9 <= shift;
      T9O <= shift(97);
    end behavioral;
    
    -- 10 ms delay
    library ieee;
    use ieee.std_logic_1164.all;
    entity srg_10 is
      port(AU_LRCK, T9O : in std_logic;
      QT10  : out std_logic_vector(97 downto 0);
      T10O : out std_logic);
    end srg_10;
  
    architecture behavioral of srg_10 is
    signal shift : std_logic_vector(97 downto 0);
    begin
    process (AU_LRCK)
    begin
      if (AU_LRCK'event and (AU_LRCK = '1')) then
        shift <= shift(96 downto 0) & T10;
      endif;
    end process;
      QT10 <= shift;
      T10O <= shift(97);
    end behavioral;
    
    
    -- =========================================
    -- FLANGER MULTIPLEXER: selects appropriately delayed data based on bouncing counter
    -- ==========================================
    library ieee;
    use ieee.std_logic_1164.all;
    entity mux_10_to_1 is
      port (BCR: in std_logic_vector(3 downto 0); -- Control logic for flanger MUX from Bouncer
            D: in std_logic_vector(1 to 10);
            FOUT: out std_logic;)
      D(1) <= T1O;
      D(2) <= T2O;
      D(3) <= T3O;
      D(4) <= T4O;
      D(5) <= T5O;
      D(6) <= T6O;
      D(7) <= T7O;
      D(8) <= T8O;
      D(9) <= T9O;
      D(10) <= T10O;  
    end mux_10_to_1;

    architecture function_table_ws of mux_10_to_1 is
    begin
      with BCR select
       FOUT <= D(1) when "0001",
               D(2) when "0010",
               D(3) when "0011",
               D(4) when "0100",
               D(5) when "0101",
               D(6) when "0110",
               D(7) when "0111",
               D(8) when "1000",
               D(9) when "1001",
               D(10) when "1010",
               'X' when others;
    end function_table_ws;

--  FULL ADDER for Flanger summation

  library IEEE;
  use IEEE.std_logic_1164.all;

  entity FBA_FLANGING is  
    port (
      S1  : in std_logic;
      FOUT : in std_logic;
      FCIN : in std_logic;
      FSUM : out std_logic;
      FCOUT: out std_logic;
      AU_LRCK: in std_logic
    );
  end FBA_FLANGING;

  architecture behavioral of FBA_FLANGING is
  FCIN <= 0;                            -- input carry starts at 0;
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      FSUM <= S1 xor FOUT xor FCIN;
      FCOUT <= (S1 and FOUT) or (S1 and FCIN) or (FOUT and FCIN);
    end if;
      FCIN <= FCOUT;
    end process;
  end behavioral;    
-- COMPLETE FLANGING UNIT



-- ECHO UNIT attached to O(2) and E(2)
  
-------------------------------------------------------------------------------
  -- ~50 ms Delay line w/ .5 Gain
  -- Delay is 2 units shorter than expected so that the LSB of each audio channel
  -- is eliminated and thus when this shifted sample is added to the original,
  -- it has the effect of adding .5 of the delayed signal to the current signal
------------------------------------------------------------------------------
  library ieee;
  use ieee.std_logic_1164.all;
  entity srg_50ms is  -- 50 ms delay
    port(AU_LRCK,  EI : in std_logic;
    QE  : out std_logic_vector(4882 downto 0);
    EO : out std_logic);
  end srg_50ms;
  
  architecture behavioral of srg_50ms is
  signal shift : std_logic_vector(4882 downto 0);
  begin
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      shift <= shift(4881 downto 0) & EI;
    endif;    
  end process;
    QE <= shift;
    EO <= .5*shift(4882);
  end behavioral;

  -- FULL ADDER TO SUM
  library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.STD_LOGIC_ARITH.ALL;

  entity FBA_ECHO is  
    port (
      EO : in std_logic;
      EI  : in std_logic;
      ECIN : in std_logic;
      ESUM : out std_logic;
      ECOUT: out std_logic;
      AU_LRCK: in std_logic
    );
  end FBA_ECHO;

  architecture behavioral of FBA_ECHO is
  ECIN <= 0; -- carry starts at zero
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      ESUM <= EI xor EO xor ECIN;
      ECOUT <= (EO and EI) or (EO and ECIN) or (EI and ECIN);
    end if;
  RCOUT <= RCIN;
  end behavioral;
-- COMPLETE ECHO UNIT


-- REVERB/VIBRATO UNIT attached between O(3) and E(3)

-- 20 ms Delay Unit w/ 5. gain on
  library ieee;
  use ieee.std_logic_1164.all;
  use IEEE.STD_LOGIC_ARITH.ALL;
  entity srg_20ms is  -- 1953 delays = 20 ms delay
    port(AU_LRCK, RSUM : in std_logic;   -- 
    QR  : out std_logic_vector(1952 downto 0);
    RO : out std_logic);
  end srg_20ms;
  
  architecture behavioral of srg_20ms is
  signal shift : std_logic_vector(    downto 0);
  begin
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      shift <= shift(1951 downto 0) & RSUM;
    endif;
  end process;
    Q <= shift;
    RO <= .5*shift(1952);
  end behavioral;

  -- FULL ADDER TO SUM
  library IEEE;
  use IEEE.std_logic_1164.all;
  use IEEE.STD_LOGIC_ARITH.ALL;
  entity FBA_RV is  
    port (
      RO : in std_logic;
      RI  : in std_logic;
      RCIN : in std_logic;
      RSUM : out std_logic;
      RCOUT: out std_logic;
      AU_LRCK: in std_logic
    );
  end FBA_RV;

  architecture behavioral of FBA_RV is
  RCIN <=0;                             -- carry starts at zero
  process (AU_LRCK)
  begin
    if (AU_LRCK'event and (AU_LRCK = '1')) then
      RSUM <= RI xor RO xor RCIN;
      RCOUT <= (RO and RI) or (RO and RCIN) or (RI and RCIN);
    end if;
  RCOUT <= RCIN;
  end behavioral;
-- COMPLETE REVERB UNIT

-------------------------------------------------------------------------------
-- NOTE: FOR REVERB AND ECHO, /2 is achieved by making the "actual size"
-- of the delay line one less than it's supposed to be, and then setting every
-- 16th bit to zero
-------------------------------------------------------------------------------
  
-- 4 to 1 MUX for end of audio effects unit
library ieee;
use ieee.std_logic_1164.all;
entity mux_4_to_1 is
  port (MUX_CONTROL: in std_logic_vector(1 downto 0); -- Control logic for MUX/DEMUX
        E: in std_logic_vector(0 to 3);
        TONE: out std_logic;)
  E(0) <= S0;
  E(1) <= FSUM;
  E(2) <= ESUM;
  E(3) <= RO;
end mux_4_to_1;

architecture function_table_ws of mux_4_to_1 is
begin
  with MUX_CONTROL select
     TONE <= E(0) when "00",            --  parallel audio out
             E(1) when "01",
             E(2) when "10",
             E(3) when "11",
             'X' when others;
end function_table_ws;
-- COMPLETE 4 to 1 MUX
